from nrresqml.factories.resqml.common import create_meta_data
from nrresqml.structures import xsd
from nrresqml.structures.energetics import EpcExternalPartReference, DataObjectReference, AbstractObject,\
    DescriptionString


def create_hdf5_reference():
    meta = create_meta_data('Hdf Proxy')
    return EpcExternalPartReference(**meta, MimeType=xsd.string('application/x-hdf5'))


def reference(obj: AbstractObject) -> DataObjectReference:
    if isinstance(obj, EpcExternalPartReference):
        title = DescriptionString('Hdf Proxy')
    else:
        title = obj.Citation.Title
    ct = xsd.string(obj.content_type_string())
    return DataObjectReference(ct, title, obj.uuid, None, None)



from dataclasses import dataclass
from typing import List, Tuple, Optional

from nrresqml.structures.resqml._rqtools import ResqmlComplexType
from nrresqml.structures import xsd
from nrresqml.structures.energetics import AxisOrder2d, AbstractVerticalCrs, AbstractProjectedCrs, LengthUom, TimeUom,\
    AbstractCitedDataObject


@dataclass
class AbstractResqmlDataObject(AbstractCitedDataObject, ResqmlComplexType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        r_ns = ResqmlComplexType.namespaces()
        a_ns = AbstractCitedDataObject.namespaces()
        order = r_ns + a_ns
        return list(dict.fromkeys(order).keys())  # Cast to dict to make sure keys are unique


@dataclass
class AbstractLocal3dCrs(AbstractResqmlDataObject):
    ArealRotation: xsd.double
    ProjectedAxisOrder: AxisOrder2d
    ProjectedUom: LengthUom
    VerticalUom: LengthUom
    XOffset: xsd.double
    YOffset: xsd.double
    ZIncreasingDownward: xsd.boolean
    ZOffset: xsd.double
    VerticalCrs: AbstractVerticalCrs
    ProjectedCrs: AbstractProjectedCrs


@dataclass
class LocalDepth3dCrs(AbstractLocal3dCrs):
    pass


@dataclass
class LocalTime3dCrs(AbstractLocal3dCrs):
    TimeUom: TimeUom



from nrresqml.structures import xsd

import nrresqml.factories.energetics
import nrresqml.factories.resqml.relationships
from nrresqml.derivatives.utils import iterate_attributes
from typing import List, Tuple, Union, Optional

from lxml import etree
from nrresqml.structures.energetics import AbstractObject, DataObjectReference


def _data_object_reference_to_element(obj: DataObjectReference, tag: str) -> etree.Element:
    return elementify(obj, [], tag)


def _create_data_object_reference(att_name: str, obj: AbstractObject, parent_ns: str) -> etree.Element:
    r = nrresqml.factories.energetics.reference(obj)
    el = _data_object_reference_to_element(r, _full_tag(parent_ns, att_name))
    return el


def _append_xsi_type(el: etree.Element, obj: Union[xsd.BasicType, xsd.XsdComplexType]):
    el.attrib[f'{{{xsd.xsi_uri}}}type'] = obj.full_type_name(abbreviate_ns=True)


def _full_tag(ns, att):
    # NB! Currently, ignore namespace for readability
    return f'{att}'


def _objectify_basic_attribute(att_name, att_obj, parent_ns) -> etree.Element:
    full_tag = _full_tag(parent_ns, att_name)
    if isinstance(att_obj, xsd.BasicType):
        ns = dict(att_obj.namespaces())
        el = etree.Element(full_tag, nsmap=ns)
    else:
        el = etree.Element(full_tag)
    if isinstance(att_obj, xsd.BasicEnum):
        el.text = att_obj.name
    else:
        el.text = str(att_obj)
    _append_xsi_type(el, att_obj)
    return el


def _attributify(att_name, att_obj) -> Tuple[str, str]:
    return att_name, str(att_obj)


def elementify(obj: xsd.XsdComplexType, existing_refs: List[AbstractObject], tag: Optional[str]) -> etree.Element:
    # el = etree.Element(obj.full_type_name(), nsmap=dict(obj.namespaces()))
    tag = tag or obj.full_type_name()
    el = etree.Element(tag, nsmap=dict(obj.namespaces()))
    for att, att_obj in iterate_attributes(obj):
        if att in obj.xml_attributes():
            key, value = _attributify(att, att_obj)
            el.attrib[key] = value
        elif isinstance(att_obj, xsd.XsdComplexType) and att_obj not in existing_refs:
            # Object is complex and needs to be recursively objectified
            el.append(elementify(att_obj, existing_refs, att))
        elif att_obj in existing_refs:
            # Object should only be addressed by reference since it already exists
            el.append(_create_data_object_reference(att, att_obj, obj.main_namespace(abbreviate=False)))
        else:
            # Object is simple and can be translated without recursion
            el.append(_objectify_basic_attribute(att, att_obj, obj.main_namespace(abbreviate=False)))
    # Inject the xsi type name
    _append_xsi_type(el, obj)
    return el



from dataclasses import dataclass

from nrresqml.structures import xsd
from nrresqml.structures.resqml.common import AbstractResqmlDataObject
from nrresqml.structures.resqml._rqtools import ResqmlBasicEnum
from nrresqml.structures.resqml.geometry import AbstractGeometry, PointGeometry


class KDirection(ResqmlBasicEnum):
    up = 0
    down = 1


class IndexableElements(ResqmlBasicEnum):
    cells = 0


@dataclass
class AbstractGridGeometry(PointGeometry):
    pass


@dataclass
class AbstractColumnLayerGridGeometry(AbstractGridGeometry):
    # Only implements k_direction. Other parameters may be required as well
    KDirection: KDirection


@dataclass
class IjkGridGeometry(AbstractColumnLayerGridGeometry):
    GridIsRightHanded: xsd.boolean


@dataclass
class AbstractRepresentation(AbstractResqmlDataObject):
    Geometry: AbstractGeometry


@dataclass
class AbstractGridRepresentation(AbstractRepresentation):
    pass


@dataclass
class AbstractColumnLayerGridRepresentation(AbstractGridRepresentation):
    Nk: xsd.positiveInteger


@dataclass
class IjkGridRepresentation(AbstractColumnLayerGridRepresentation):
    Ni: xsd.positiveInteger
    Nj: xsd.positiveInteger



from dataclasses import dataclass
from typing import List, Tuple, Optional

from nrresqml.structures import xsd


@dataclass
class _ContentTypesComplexType(xsd.XsdComplexType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [(None, 'http://schemas.openxmlformats.org/package/2006/content-types')]


@dataclass
class Default(_ContentTypesComplexType):
    ContentType: str
    Extension: str

    @classmethod
    def xml_attributes(cls) -> List[str]:
        return ['ContentType', 'Extension'] + super().xml_attributes()


@dataclass
class Override(_ContentTypesComplexType):
    ContentType: str
    PartName: str

    @classmethod
    def xml_attributes(cls) -> List[str]:
        return ['ContentType', 'PartName'] + super().xml_attributes()


@dataclass
class Types(_ContentTypesComplexType):
    Default: List[Default]
    Override: List[Override]



from typing import Dict

from nrresqml.factories.resqml.common import create_meta_data
from nrresqml.structures import xsd
from nrresqml.structures.energetics import EpcExternalPartReference, Hdf5Dataset
from nrresqml.structures.resqml import properties as ps
from nrresqml.structures.resqml.geometry import DoubleHdf5Array, IntegerHdf5Array
from nrresqml.structures.resqml.properties import StringTableLookup, StringLookup
from nrresqml.structures.resqml.representations import AbstractRepresentation, IndexableElements


def create_continuous_property(title: str, h5_path: str, min_value: float, max_value: float,
                               rep: AbstractRepresentation, h5p: EpcExternalPartReference,) -> ps.ContinuousProperty:
    """
    Creates a continuous property that is related to some other supporting representation. Convenience factory function
    that fills in default values where appropriate. These can be overwritten post-construction
    """
    meta = create_meta_data(title)
    a_kind = ps.StandardPropertyKind(ps.ResqmlPropertyKind.volume_per_volume)
    h5d = Hdf5Dataset(xsd.string(h5_path), h5p)
    ava = DoubleHdf5Array(h5d)
    pov = ps.PatchOfValues(xsd.integer(0), ava)
    prop = ps.ContinuousProperty(
        IndexableElement=IndexableElements.cells,
        Count=xsd.positiveInteger(1),
        RealizationIndex=None,
        TimeStep=None,
        SupportingRepresentation=rep,
        PropertyKind=a_kind,
        PatchOfValues=pov,
        MinimumValue=xsd.double(min_value),
        MaximumValue=xsd.double(max_value),
        Uom=ps.ResqmlUom.m,
        **meta
    )
    return prop


def create_categorical_property(title: str, h5_path: str, rep: AbstractRepresentation, h5p: EpcExternalPartReference,
                                value_map: Dict[int, str]) -> ps.CategoricalProperty:
    meta = create_meta_data(title)
    a_kind = ps.StandardPropertyKind(ps.ResqmlPropertyKind.categorical)
    h5d = Hdf5Dataset(xsd.string(h5_path), h5p)
    assert -1 not in value_map  # Not supported
    ava = IntegerHdf5Array(h5d, xsd.integer(-1))
    pov = ps.PatchOfValues(xsd.integer(0), ava)
    # Look-up table
    tab_meta = create_meta_data(f'{title} - look-up')
    values = [
        StringLookup(xsd.integer(key), xsd.string(value))
        for key, value in value_map.items()
    ]
    tab = StringTableLookup(Value=values, **tab_meta)
    prop = ps.CategoricalProperty(
        IndexableElement=IndexableElements.cells,
        Count=xsd.positiveInteger(1),
        RealizationIndex=None,
        TimeStep=None,
        SupportingRepresentation=rep,
        PropertyKind=a_kind,
        PatchOfValues=pov,
        Lookup=tab,
        **meta
    )
    return prop



from nrresqml.factories.resqml.common import create_meta_data
from nrresqml.structures import xsd
from nrresqml.structures.energetics import VerticalUnknownCrs, ProjectedCrsEpsgCode, AxisOrder2d, LengthUom
from nrresqml.structures.resqml.common import LocalDepth3dCrs
from nrresqml.structures.resqml.geometry import Point3dOffset, Point3d, Point3dLatticeArray
from nrresqml.structures.resqml.representations import IjkGridGeometry, KDirection, IjkGridRepresentation


def create_local_depth_3d_crs():
    v_crs = VerticalUnknownCrs(xsd.string('Unknown'))
    p_crs = ProjectedCrsEpsgCode(xsd.positiveInteger(4146))
    crs_meta = create_meta_data('Delft 3D CRS')
    crs = LocalDepth3dCrs(
        **crs_meta,
        ArealRotation=xsd.double(0.0),
        ProjectedAxisOrder=AxisOrder2d.easting_northing,
        ProjectedUom=LengthUom.m,
        VerticalUom=LengthUom.m,
        XOffset=xsd.double(0.0),
        YOffset=xsd.double(0.0),
        ZIncreasingDownward=xsd.boolean(True),
        ZOffset=xsd.double(0.0),
        VerticalCrs=v_crs,
        ProjectedCrs=p_crs
    )
    return crs


def create_grid_representation_for_regular_grid(ni, nj, nk, dx, dy, dz) -> IjkGridRepresentation:
    crs = create_local_depth_3d_crs()
    ds = Point3dOffset(Point3d(dx, dy, dz))
    arr = Point3dLatticeArray(xsd.boolean(True), Point3d(xsd.double(0.0), xsd.double(0.0), xsd.double(0.0)), ds)
    geom = IjkGridGeometry(crs, arr, KDirection.down, xsd.boolean(True))
    grid_meta = create_meta_data('Converted Delft 3D grid')
    grid = IjkGridRepresentation(**grid_meta, Geometry=geom, Ni=ni, Nj=nj, Nk=nk)
    return grid



from dataclasses import dataclass
from typing import List, Tuple, Optional

from nrresqml.structures import xsd


class W3CDTF(xsd.dateTime):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [('dcterms', 'http://purl.org/dc/terms/')] + super().namespaces()


@dataclass
class coreProperties(xsd.XsdComplexType):
    # TODO: the namespace of these types is not correct
    created: W3CDTF
    creator: xsd.string

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [
                   ('cp', 'http://schemas.openxmlformats.org/package/2006/metadata/core-properties'),
                   ('dc', 'http://purl.org/dc/elements/1.1/'),
                   ('dcterms', 'http://purl.org/dc/terms/'),
               ] + super().namespaces()



import os
import pathlib
import shutil
from zipfile import ZipFile

from lxml import etree

from nrresqml.serialization import elementify
from nrresqml.derivatives.hdf5resqmladaptor import Hdf5ResQmlAdaptor
from nrresqml.resqml import ResQml
from nrresqml.structures import contenttypes, xsd
from nrresqml.structures.energetics import EpcExternalPartReference, AbstractObject
from nrresqml.structures.relationships import Relationships, Relationship


_RELS_DIR = '_rels'


def _create_cache_folder(cache_dir: pathlib.Path):
    os.makedirs(str(cache_dir), exist_ok=False)
    os.makedirs(str(cache_dir / _RELS_DIR))
    os.makedirs(str(cache_dir / 'docProps'))


class _Cacher:
    def __init__(self, save_path: pathlib.Path, zip_mode: bool) -> None:
        self._zip_mode = zip_mode
        self._epc_file_path = save_path.with_suffix('.epc')
        # Initialize folders
        if not zip_mode:
            # Will throw if folder already exists
            _create_cache_folder(self._epc_file_path)
        else:
            # Ok if folder exists, but file cannot
            os.makedirs(str(self._epc_file_path.parent), exist_ok=True)
            if self._epc_file_path.is_file():
                os.remove(str(self._epc_file_path))

    def _handle(self, base_name: str):
        if self._zip_mode:
            return ZipFile(self._epc_file_path, 'a').open(base_name, 'w')
        else:
            return str(self._epc_file_path / base_name)

    def dump_epc_object(self, obj: AbstractObject, el: etree.Element):
        tree_obj = etree.ElementTree(el)
        base_name = obj.base_name()
        fh = self._handle(base_name)
        tree_obj.write(fh, pretty_print=True, encoding='utf-8', xml_declaration=True)

    def dump_relationships(self, obj: AbstractObject, rels: Relationships):
        el = elementify.elementify(rels, [], None)
        tree_obj = etree.ElementTree(el)
        fn = pathlib.Path(_RELS_DIR) / (obj.base_name() + '.rels')
        fh = self._handle(str(fn))
        tree_obj.write(fh, pretty_print=True, standalone=False, encoding='utf-8')

    def dump_content_types(self, ct: contenttypes.Types):
        el = elementify.elementify(ct, [], None)
        tree = etree.ElementTree(el)
        fh = self._handle('[ContentTypes].xml')
        tree.write(fh, pretty_print=True, standalone=False, encoding='utf-8')

    def dump_dot_rels(self):
        type_ = 'http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties'
        r = [Relationship('rId1', 'docProps/core.xml', None, type_)]
        rs = Relationships(r)
        el = elementify.elementify(rs, [], None)
        tree = etree.ElementTree(el)
        fh = self._handle('_rels/.rels')
        tree.write(fh, pretty_print=True, standalone=False, encoding='utf-8')

    def dump_core(self):
        from nrresqml.structures import core
        core = core.coreProperties(core.W3CDTF.now(), xsd.string('NR ResQml from NetCDF'))
        el = elementify.elementify(core, [], None)
        tree = etree.ElementTree(el)
        fh = self._handle('docProps/core.xml')
        tree.write(fh, pretty_print=True, standalone=False, encoding='utf-8')


def build_from_adaptor(adaptor: Hdf5ResQmlAdaptor, save_path: pathlib.Path, use_zip=False) -> ResQml:
    """
    Creates a ResQml object from the given adaptor and writes it to file. Returns the created ResQml instance for
    convenience
    """
    # Create the objects
    obs = adaptor.create_objects()
    h5_fn = save_path.with_suffix('.h5').name
    rq = ResQml(obs, save_path)
    for ec in rq.objects(EpcExternalPartReference):
        rq.set_hdf5_reference(ec, h5_fn)

    # Set up the cacher
    c = _Cacher(save_path.with_suffix('.epc'), use_zip)

    # Dump objects
    for obj in obs:
        el = elementify.elementify(obj, obs, None)
        c.dump_epc_object(obj, el)

    # Dump relationships
    for obj in obs:
        rs = rq.relationships(obj)
        c.dump_relationships(obj, rs)

    # Dump content types
    c.dump_content_types(rq.content_types())

    # Dump datafile
    df = save_path.parent / h5_fn
    adaptor.dump_h5_file(df)

    # Dump boiler-plate files
    c.dump_dot_rels()
    c.dump_core()

    return rq



import glob
import pathlib
from dataclasses import is_dataclass
from typing import List, Dict, Iterator, Optional, Type
from zipfile import ZipFile

from lxml import etree

from nrresqml.serialization.postprocessing import resolve_references
from nrresqml.serialization import objectify
from nrresqml.derivatives import utils
from nrresqml.factories.resqml import relationships
from nrresqml.structures import contenttypes
from nrresqml.structures.energetics import AbstractObject, EpcExternalPartReference, UuidString
from nrresqml.structures.relationships import Relationships


def _create_content_type_override(obj: AbstractObject) -> contenttypes.Override:
    return contenttypes.Override(obj.content_type_string(), '/' + obj.base_name())


class ResQml:
    def __init__(self, objects: List[AbstractObject], root_path: pathlib.Path) -> None:
        assert root_path.suffix == '.epc'
        # Path to the root directory of the ResQml file. Primarily used to access the HDF5 data file properly
        self._root_path = root_path
        # List of all the ResQml objects (internal representations)
        self._objects = objects
        # Dict that maps uuid's to their corresponding HDF5 file. For the ResQml object to be consistent, it is
        # necessary that each object of type EpcExternalPartReference has a corresponding entry in this dict.
        self._hdf5_refs: Dict[UuidString, str] = {}

    def set_hdf5_reference(self, obj: EpcExternalPartReference, hdf5_file: str):
        assert obj in self._objects
        self._hdf5_refs[obj.uuid] = hdf5_file

    def get_full_hdf5_reference(self, obj: EpcExternalPartReference) -> str:
        return str(self._root_path.parent / self._hdf5_refs[obj.uuid])

    def objects(self, types: Optional[Type] = None) -> Iterator[AbstractObject]:
        for obj in self._objects:
            if types is None or isinstance(obj, types):
                yield obj

    def find(self, uuid: UuidString) -> Optional[AbstractObject]:
        for o in self._objects:
            if o.uuid == uuid:
                return o
        return None

    def relationships(self, obj: AbstractObject) -> Relationships:
        # Check if obj is an EpcExternalPartReference, in which case it is handled in a special way
        if isinstance(obj, EpcExternalPartReference):
            fp = self._hdf5_refs[obj.uuid]
            rels = relationships.create_epc_external_part_reference_relationship(obj, pathlib.Path(fp))
            return Relationships([rels])

        r = []
        for _, att_obj in utils.iterate_attributes(obj):
            if att_obj in self._objects:
                # Attribute is referenced, create and append relationship
                r.append(relationships.create_general_relationship(att_obj))
            elif is_dataclass(att_obj):
                # Attribute is not referenced, but may reference other objects deeper in the hierarchy
                r += self.relationships(att_obj).Relationship
        return Relationships(r)

    def content_types(self) -> contenttypes.Types:
        defs = [
            contenttypes.Default('application/vnd.openxmlformats-package.relationships+xml', 'rels'),
            contenttypes.Default('application/xml', 'xml'),
        ]
        overrides = [_create_content_type_override(obj) for obj in self._objects]
        return contenttypes.Types(defs, overrides)

    @staticmethod
    def read(cache_dir: pathlib.Path) -> 'ResQml':
        # Load objects
        fn = glob.glob(str(cache_dir) + '/*.xml')
        ets = [etree.parse(f) for f in fn]
        objs = [objectify.objectify(e.getroot()) for e in ets]
        [resolve_references(obj, objs) for obj in objs]
        rq = ResQml(objs, cache_dir)

        # Set hdf5 references
        h5_file = cache_dir.with_suffix('.h5').name  # Assumed convention to ensure relative paths are correct
        for epr in rq.objects(EpcExternalPartReference):
            rq.set_hdf5_reference(epr, h5_file)
        return rq

    @staticmethod
    def read_zipped(epc_file: pathlib.Path) -> 'ResQml':
        with ZipFile(epc_file, 'r') as z:
            ets = [
                etree.parse(z.open(n))
                for n in z.namelist()
                if n.endswith('.xml') and n.startswith('obj_')
            ]
            objs = [objectify.objectify(e.getroot()) for e in ets]
            [resolve_references(obj, objs) for obj in objs]
            rq = ResQml(objs, epc_file)

            # Set hdf5 references
            h5_file = epc_file.with_suffix('.h5').name  # Assumed convention to ensure relative paths are correct
            for epr in rq.objects(EpcExternalPartReference):
                rq.set_hdf5_reference(epr, h5_file)
        return rq



def iterate_attributes(obj):
    for att in dir(obj):
        if att.startswith('_'):
            continue
        att_obj = getattr(obj, att)
        if att_obj is None:
            continue
        if callable(att_obj):
            continue
        if isinstance(att_obj, list):
            for itr in att_obj:
                yield att, itr
        else:
            yield att, att_obj



from dataclasses import is_dataclass
from typing import List
from nrresqml.structures.energetics import AbstractObject, DataObjectReference
from nrresqml import reporting


def resolve_references(obj, refs: List[AbstractObject]):
    for an in dir(obj):
        if an.startswith('_'):
            continue
        ao = getattr(obj, an)
        if isinstance(ao, DataObjectReference):
            ds = [r for r in refs if r.uuid == ao.UUID]
            if len(ds) == 0:
                reporting.error(f'A DataObjectReference is referring a non-existing object ({ao.UUID})')
                continue
            if len(ds) > 1:
                reporting.error(f'A DataObjectReference is referring a duplicate uuid ({ao.UUID})')
                continue
            d = ds[0]
            setattr(obj, an, d)
            resolve_references(d, refs)
        elif is_dataclass(ao):
            resolve_references(ao, refs)



from dataclasses import dataclass

import h5py
import numpy as np

from nrresqml.factories.resqml.common import create_meta_data
from nrresqml.factories.resqml.representations import create_local_depth_3d_crs
from nrresqml.structures import xsd
from nrresqml.structures.energetics import EpcExternalPartReference, Hdf5Dataset
from nrresqml.structures.resqml.geometry import DoubleHdf5Array, Point3dHdf5Array, ParametricLineArray,\
    Point3dParametricArray
from nrresqml.structures.resqml.representations import IjkGridRepresentation, IjkGridGeometry, KDirection


class IjkGridCreationError(Exception):
    pass


@dataclass
class _GridParameters:
    nx: int
    ny: int
    nz: int
    dx: float
    dy: float
    zz: np.ndarray
    x0: float
    y0: float


def _extract_grid_parameters(_d3_file: h5py.File) -> _GridParameters:
    # Check that the file contains the expected parameters
    _keys = set(_d3_file.keys())
    if not {'XCOR', 'xcor'}.intersection(_keys):
        raise IjkGridCreationError('Missing at least parameter "XCOR" or "xcor"')
    if not {'YCOR', 'ycor'}.intersection(_keys):
        raise IjkGridCreationError('Missing at least parameter "YCOR" or "ycor"')
    if not {'DPS', 'zcor'}.intersection(_keys):
        raise IjkGridCreationError('Missing at least parameter "DPS" or "zcor"')
    # Extract parameters
    xcor = _d3_file['xcor'] if 'xcor' in _keys else _d3_file['XCOR']
    ycor = _d3_file['ycor'] if 'ycor' in _keys else _d3_file['YCOR']
    zcor = _d3_file['zcor'] if 'zcor' in _keys else _d3_file['DPS']
    nk, ni, nj = zcor.shape
    dx = (np.max(xcor) - np.min(xcor)) / (ni - 1)
    dy = (np.max(ycor) - np.min(ycor)) / (nj - 1)
    zz = np.array(zcor)
    if 'zcor' not in _d3_file:
        # DPS was used, which is a temporal layer and may not be spatially feasible
        zz = IjkGridCreator.mono_elevation(zz)
    return _GridParameters(ni, nj, nk, dx, dy, zz, xcor[0, 0], ycor[0, 0])


def _hdf5_array(h5_epc_ref: EpcExternalPartReference, h5_path: str):
    vals = Hdf5Dataset(xsd.string(h5_path), h5_epc_ref)
    arr = DoubleHdf5Array(vals)
    return arr


class IjkGridCreator:
    def __init__(self, d3_file: h5py.File) -> None:
        """
        Class used to create an IJK grid from a given data file. The file is expected to come from Delft 3D. This means
        that it is assumed to contain at least the variables DPS, XCOR and YCOR, which will be used to build the grid.
        Alternatively, zcor, xcor, ycor. Most importantly, it interprets the grid to have flat cell tops.

        :param d3_file: A HDF5 file, imported by h5py
        """
        # Grid geometry data
        gp = _extract_grid_parameters(d3_file)
        self._control_points = IjkGridCreator.pillarized_control_points(gp)
        # Control point parameters. Describes each pillar as monotonized z values
        self._control_point_parameters = IjkGridCreator.pillarize(gp.zz)

    @property
    def _control_points_path(self):
        return f'control_points_{id(self._control_points)}'

    @property
    def _control_point_parameters_path(self):
        return f'control_point_parameters_{id(self._control_point_parameters)}'

    @property
    def pillars(self) -> np.ndarray:
        """
        Returns the regularly spaced pillars that represents the grid as a Nx x Ny x Nz grid. The pillars are
        monotonically increasing in the z-direction
        :return:
        """
        # With the current implementation, the pillars exactly match the control point parameters.
        return self._control_point_parameters

    def ijk_representation(self, h5_epc_ref: EpcExternalPartReference) -> IjkGridRepresentation:
        """
        Creates and returns the IjkGridRepresentation for the input data file

        :param h5_epc_ref: The EpcExternalPartReference that is used to reference the data file (HDF5 file) that
                           corresponds to the ResQml file.

        :return:           The IjkGridRepresentation of the grid constructed from the d3_file
        """
        # Create geometry
        crs = create_local_depth_3d_crs()
        plane = Point3dHdf5Array(Hdf5Dataset(xsd.string(self._control_points_path), h5_epc_ref))
        lines = ParametricLineArray(
            _hdf5_array(h5_epc_ref, self._control_point_parameters_path),
            plane
        )
        points = Point3dParametricArray(_hdf5_array(h5_epc_ref, self._control_points_path), lines)
        geom = IjkGridGeometry(crs, points, KDirection.up, xsd.boolean(True))
        # Create representation
        meta = create_meta_data('Delft 3D-based grid')
        rep = IjkGridRepresentation(
            **meta,
            Geometry=geom,
            Nk=xsd.positiveInteger(self._control_point_parameters.shape[3]),
            Ni=xsd.positiveInteger(self._control_point_parameters.shape[1]),
            Nj=xsd.positiveInteger(self._control_point_parameters.shape[2])
        )
        return rep

    def dump_grid_data(self, h5_file: h5py.File):
        h5_file.create_dataset(
            self._control_points_path, data=self._control_points, compression='gzip'
        )
        h5_file.create_dataset(
            self._control_point_parameters_path, data=self._control_point_parameters, compression='gzip'
        )

    @staticmethod
    def mono_elevation(depth: np.ndarray) -> np.ndarray:
        """
        Convert depth values to elevation values, and truncate them at
        every time step to ensure elevation is monotonically nondecreasing
        in time at every grid location.
        :param depth: DPS from Delft3D output file. 3D numpy array
        :return: truncated elevation values. 3D numpy array
        """
        base_depth = depth[0, :, :]
        relative_elevation = depth - base_depth
        truncated_elevation = np.maximum.accumulate(relative_elevation[::-1, :, :], axis=0)[::-1]
        return -truncated_elevation

    @staticmethod
    def pillarized_control_points(gp: _GridParameters):
        cp = np.ndarray((4, gp.nx, gp.ny, 3), dtype=np.float)
        xx, yy = np.meshgrid(np.arange(gp.nx + 1) * gp.dx, np.arange(gp.ny + 1) * gp.dy, indexing='ij')
        # Control points. Describes the lateral distribution of pillars
        cp[0, :, :, 0] = xx[:-1, :-1]
        cp[1, :, :, 0] = xx[1:, :-1]
        cp[2, :, :, 0] = xx[1:, 1:]
        cp[3, :, :, 0] = xx[-1:, 1:]
        cp[0, :, :, 1] = yy[:-1, :-1]
        cp[1, :, :, 1] = yy[1:, :-1]
        cp[2, :, :, 1] = yy[1:, 1:]
        cp[3, :, :, 1] = yy[-1:, 1:]

        # Add reference and set z-value to nan
        cp[:, :, :, 0] += gp.x0
        cp[:, :, :, 1] += gp.y0
        cp[:, :, :, 2] = np.nan
        return cp

    @staticmethod
    def pillarize(pillars: np.ndarray) -> np.ndarray:
        """
        Convert shared pillars (nz x nx x ny) to split pillars (4 x nx x ny x nz)
        """
        assert pillars.ndim == 3
        nk, ni, nj = pillars.shape
        column_pillars = np.zeros((4, ni, nj, nk), dtype=pillars.dtype)
        pt = pillars.transpose((1, 2, 0))
        # Duplicate pillars.
        column_pillars[0, :, :, :] = pt
        column_pillars[1, :, :, :] = pt
        column_pillars[2, :, :, :] = pt
        column_pillars[3, :, :, :] = pt
        return column_pillars



import pathlib

from nrresqml.structures.energetics import EpcExternalPartReference, AbstractObject
from nrresqml.structures.relationships import Relationship, TargetMode


def create_epc_external_part_reference_relationship(obj: EpcExternalPartReference, file_path: pathlib.Path
                                                    ) -> Relationship:
    return Relationship(str(obj.uuid), str(file_path), TargetMode.External,
                        'http://schemas.energistics.org/package/2012/relationships/externalResource')


def create_general_relationship(obj: AbstractObject) -> Relationship:
    return Relationship(str(obj.uuid), obj.base_name(), None,
                        'http://schemas.energistics.org/package/2012/relationships/externalPartProxyToMl')



import sys
import os
import pathlib

from nrresqml.derivatives.hdf5resqmladaptor import Delft3DResQmlAdaptor, AdaptorError
from nrresqml.derivatives import rqbuilder as rio


def _derive_archel_name(delft3d_name: str):
    # TODO: handle the variety of naming conventions for files containing architectural elements/subenvironments.
    #  The following observations are made:
    #    * Subenvironment files usually exists for older results, but are less refined
    #    * Architectural elements files does not yet exist on the threadds server
    #    * Architectural elements files seem to contain both subenvironment and architectural elements data sets
    if delft3d_name.startswith('http'):
        # Files on the OpenDAP server does not yet contain architectural elements, only sub-environment
        return os.path.dirname(delft3d_name).replace('simulation', 'postprocess') + '/subenvironment.nc'
    else:
        an = os.path.dirname(delft3d_name) + '/architectural_elements.nc'
        if os.path.isfile(an):
            return an
        else:
            # This should perhaps be the first attempt, if the file has the required keys (archel/subenv)
            return delft3d_name


def convert_delft3d_to_resqml(delft3d_file_name: str, resqml_output_directory: str) -> None:
    """
    Converts an existing Delft3D NetCdf file to a ResQml file(s)

    :param delft3d_file_name:       File path to the NetCdf file (usually a .nc file)
    :param resqml_output_directory: Path to an existing directory where the ResQml database should be written. The
                                    directory must already exist and be writable. The following two files will be
                                    written:
                                        - resqml_output_directory/<resqml_file_name>.epc
                                        - resqml_output_directory/<resqml_file_name>.h5
                                     The base name 'resqml_file_name' is derived from delft3d_file_name
    """
    try:
        # Derive name of file containing architectural elements from the base name
        archel_file = _derive_archel_name(delft3d_file_name)
        daf = Delft3DResQmlAdaptor(
            delft3d_file_name,
            archel_file,
        )
    except AdaptorError as e:
        print('Failed to create ResQml database:')
        print(e)
        sys.exit(1)
    save_path = pathlib.Path(resqml_output_directory) / pathlib.Path(delft3d_file_name).with_suffix('.epc').name
    rio.build_from_adaptor(daf, save_path, True)



from dataclasses import dataclass
from typing import Optional

from nrresqml.structures.energetics import Hdf5Dataset
from nrresqml.structures import xsd
from nrresqml.structures.resqml.common import AbstractLocal3dCrs
from nrresqml.structures.resqml._rqtools import ResqmlComplexType


@dataclass
class Point3d(ResqmlComplexType):
    Coordinate1: xsd.double
    Coordinate2: xsd.double
    Coordinate3: xsd.double


@dataclass
class Point3dOffset(ResqmlComplexType):
    Offset: Point3d


""" Abstract Value Array """


@dataclass
class AbstractValueArray(ResqmlComplexType):
    pass


@dataclass
class AbstractIntegerArray(AbstractValueArray):
    pass


@dataclass
class IntegerHdf5Array(AbstractIntegerArray):
    Values: Hdf5Dataset
    NullValue: xsd.integer


@dataclass
class AbstractDoubleArray(AbstractValueArray):
    pass


@dataclass
class DoubleHdf5Array(AbstractDoubleArray):
    Values: Hdf5Dataset


""" AbstractPoint3dArray """


@dataclass
class AbstractPoint3dArray(ResqmlComplexType):
    pass


@dataclass
class Point3dLatticeArray(AbstractPoint3dArray):
    AllDimensionsAreOrthogonal: Optional[xsd.boolean]
    Origin: Point3d
    Offset: Point3dOffset


@dataclass
class Point3dHdf5Array(AbstractPoint3dArray):
    Coordinates: Hdf5Dataset


@dataclass
class AbstractGeometry(ResqmlComplexType):
    LocalCrs: AbstractLocal3dCrs


@dataclass
class PointGeometry(AbstractGeometry):
    Points: AbstractPoint3dArray


@dataclass
class AbstractPoint3dArray(ResqmlComplexType):
    pass


@dataclass
class AbstractParametricLineArray(ResqmlComplexType):
    pass


@dataclass
class ParametricLineArray(AbstractParametricLineArray):
    ControlPointParameters: AbstractDoubleArray
    ControlPoints: AbstractPoint3dArray


@dataclass
class Point3dParametricArray(AbstractPoint3dArray):
    Parameters: AbstractValueArray
    ParametricLines: AbstractParametricLineArray



import numpy as np
import h5py
from typing import Optional

from nrresqml.resqml import ResQml
from nrresqml.structures.energetics import Hdf5Dataset
from nrresqml.structures.resqml.geometry import IntegerHdf5Array, DoubleHdf5Array, Point3dParametricArray, \
    ParametricLineArray, Point3dHdf5Array
from nrresqml.structures.resqml.properties import CategoricalProperty, ContinuousProperty
from nrresqml.structures.resqml.representations import AbstractRepresentation, IjkGridRepresentation, IjkGridGeometry


def crop_array(arr_zxy,
               array_x0,
               array_y0,
               array_dx,
               array_dy,
               crop_x0,
               crop_y0,
               crop_x1,
               crop_y1):
    assert crop_x1 > crop_x0 and crop_y1 > crop_y0
    nx, ny = arr_zxy.shape[1:]
    # Determine crop indexes
    i0 = np.ceil((crop_x0 - array_x0) / array_dx)
    j0 = np.ceil((crop_y0 - array_y0) / array_dy)
    i1 = np.floor((crop_x1 - array_x0) / array_dx)
    j1 = np.floor((crop_y1 - array_y0) / array_dy)
    i0 = int(max(0, i0))
    j0 = int(max(0, j0))
    i1 = int(min(nx, i1))
    j1 = int(min(ny, j1))
    assert j1 > j0 and i1 > i0  # Verify that the crop box is properly specified
    return arr_zxy[:, i0:i1, j0:j1]


def extract_geometry(rq: ResQml, flatten_pillars: bool, indexing: str):
    assert indexing in ('ijk', 'kij')
    # Extract pillars and other grid parameters
    z = list(rq.objects(IjkGridRepresentation))
    assert len(z) == 1  # Should be one and only one grid in the object. Multiple grids are not supported (yet)
    # Assert on types to be explicit about the assumed types and to aid code completion
    ijk = z[0]
    assert isinstance(ijk, IjkGridRepresentation)
    assert isinstance(ijk.Geometry, IjkGridGeometry)
    assert isinstance(ijk.Geometry.Points, Point3dParametricArray)
    assert isinstance(ijk.Geometry.Points.ParametricLines, ParametricLineArray)
    assert isinstance(ijk.Geometry.Points.ParametricLines.ControlPointParameters, DoubleHdf5Array)
    pillars = _extract_dataset(rq, ijk.Geometry.Points.ParametricLines.ControlPointParameters.Values)
    assert isinstance(ijk.Geometry.Points.ParametricLines.ControlPoints, Point3dHdf5Array)
    xxyyzz = _extract_dataset(rq, ijk.Geometry.Points.ParametricLines.ControlPoints.Coordinates)
    xx, yy = xxyyzz[:, :, :, 0], xxyyzz[:, :, :, 1]
    if flatten_pillars:
        xx = xx[0, :, :]
        yy = yy[0, :, :]
        pillars = np.mean(pillars, axis=0)
    pillars = pillars.transpose({'ijk': (0, 1, 2), 'kij': (2, 0, 1)}[indexing])
    return ijk, xx, yy, pillars


def extract_property(resqml: ResQml, supp_rep: Optional[AbstractRepresentation], h5_path: str, categorical: bool):
    if categorical:
        p_type = CategoricalProperty
        a_type = IntegerHdf5Array
    else:
        p_type = ContinuousProperty
        a_type = DoubleHdf5Array
    props = [
        a for a in resqml.objects(p_type)
        if isinstance(a, p_type)
        and isinstance(a.PatchOfValues.Values, a_type)
        and a.PatchOfValues.Values.Values.PathInHdfFile == h5_path
        and (supp_rep is None or a.SupportingRepresentation.uuid == supp_rep.uuid)
    ]
    assert len(props) == 1
    prop = props[0].PatchOfValues.Values
    assert isinstance(prop, a_type)
    return _extract_dataset(resqml, prop.Values)


def _extract_dataset(resqml: ResQml, hdf5_dataset: Hdf5Dataset):
    hdf5_path = resqml.get_full_hdf5_reference(hdf5_dataset.HdfProxy)
    h5ds = h5py.File(hdf5_path, mode='r')
    hdf5_path = hdf5_dataset.PathInHdfFile
    return h5ds[hdf5_path]



import pathlib
import pydap.client
import webob.exc
from pydap.model import DatasetType
from typing import List, Union

import h5py
import numpy as np

from nrresqml.derivatives.ijkgridcreator import IjkGridCreator, IjkGridCreationError
from nrresqml.factories.energetics import create_hdf5_reference
from nrresqml.factories.resqml.properties import create_continuous_property, create_categorical_property
from nrresqml.structures.energetics import AbstractObject


class AdaptorError(Exception):
    pass


class Hdf5ResQmlAdaptor:
    """
    Interface class for converting arbitrary data into an hdf5 file suited for ResQml, and its corresponding ResQml data
    objects
    """

    def create_objects(self) -> List[AbstractObject]:
        raise NotImplementedError('Method not implemented by subclass')

    def dump_h5_file(self, filename: pathlib.Path):
        raise NotImplementedError('Method not implemented by subclass')

    def h5_base_name(self) -> str:
        raise NotImplementedError('Method not implemented by subclass')


def _open_delft3d_path(p: str) -> Union[h5py.File, DatasetType]:
    if p.startswith('http'):
        return pydap.client.open_url(p)
    else:
        return h5py.File(p, mode='r')


class Delft3DResQmlAdaptor(Hdf5ResQmlAdaptor):
    # TODO: implement proper mapping and move definitions somewhere else
    _archel_map = {
        0: 'Inactive',
        1: 'Active Channel',
        2: 'Channel Fill',
        3: 'Delta Top Background',
        4: 'Mouth Bar',
        5: 'Delta Front Background',
        6: 'Pro-delta',
    }
    _subenviron_map = {
        0: 'Background',
        1: 'Delta Top',
        2: 'Delta Front',
        3: 'Pro-delta',
    }
    # HDF5 keys
    _delft3d_archel_key = 'archel'
    _delft3d_subenv_key = 'subenvironment'
    _resqml_archel_key = 'archel'
    _resqml_subenv_key = 'subenvironment'
    # Continuous attributes and their natural bounds. We can compute the bounds per data set, but it is not obvious from
    # the ResQML specification that it is required. Instead, we choose to use some natural bounds based on what the
    # properties represents. This can save a significant amount of computation time.
    _cont_prop_bounds = {
        'DXX01': (0.0, 50.0),
        'DXX02': (0.0, 50.0),
        'DXX03': (0.0, 50.0),
        'DXX04': (0.0, 50.0),
        'DXX05': (0.0, 50.0),
        'd50_per_sedclass': (0.0, 50.0),
        'diameter': (0.0, 50.0),
        'fraction': (0.0, 1.0),
        'sorting': (0.0, 1.0),
        'Sed1_mass': (0.0, 1e6),
        'Sed2_mass': (0.0, 1e6),
        'Sed3_mass': (0.0, 1e6),
        'Sed4_mass': (0.0, 1e6),
        'Sed5_mass': (0.0, 1e6),
        'Sed6_mass': (0.0, 1e6),
        'Sed1_volfrac': (0.0, 1.0),
        'Sed2_volfrac': (0.0, 1.0),
        'Sed3_volfrac': (0.0, 1.0),
        'Sed4_volfrac': (0.0, 1.0),
        'Sed5_volfrac': (0.0, 1.0),
        'Sed6_volfrac': (0.0, 1.0),
        'Porosity': (0.0, 1.0),
        'porosity': (0.0, 1.0),
        'permeability': (0.0, 1.0),
    }

    def __init__(self, d3_file: str, archel_file: str) -> None:
        self._d3_file = _open_delft3d_path(d3_file)
        try:
            self._archel_file = _open_delft3d_path(archel_file)
        except (webob.exc.HTTPError, OSError):
            print(f'Architectural elements/sub-environment not defined: Failed to open file {archel_file}')
            self._archel_file = None

        # Grid handling
        try:
            self._grid_creator = IjkGridCreator(self._d3_file)
        except IjkGridCreationError as e:
            raise AdaptorError(f'Failed to create grid from {d3_file} with the following error:\n  ' + str(e))

        # Continuous properties
        self._continuous_properties = []
        for pn in self._cont_prop_bounds:
            try:
                self._continuous_properties.append(self._d3_file[pn])
                print(f'Included continuous property {pn}')
            except KeyError:
                print(f'Did not find property {pn}')

    def create_objects(self) -> List[AbstractObject]:
        ref = create_hdf5_reference()
        ijk = self._grid_creator.ijk_representation(ref)
        # Create continuous properties
        props = []
        for p in self._continuous_properties:
            # Workaround to support both HDF5 and pydap when finding 'long_name':
            try:
                long_name = p.attrs['long_name'].decode('utf-8')
            except AttributeError:
                long_name = p.attributes['long_name']
            pn = p.name.strip('/')
            props.append(create_continuous_property(
                long_name,
                pn,
                self._cont_prop_bounds[pn][0],
                self._cont_prop_bounds[pn][1],
                ijk,
                ref
            ))
        # Create categorical properties
        # Architectural elements
        props.append(create_categorical_property('Architectural element', self._resqml_archel_key, ijk, ref,
                                                 self._archel_map))
        # Sub-environment
        props.append(create_categorical_property('Subenvironment', self._resqml_subenv_key, ijk, ref,
                                                 self._subenviron_map))

        return [ijk, ijk.Geometry.LocalCrs, ref] + props

    def dump_h5_file(self, filename: pathlib.Path):
        out = h5py.File(filename, 'w')
        for p in self._continuous_properties:
            out.copy(p, p.name)

        # Define temporary function to extract archel data
        def _copy_archel_data(source, target):
            if self._archel_file is None or source not in self._archel_file:
                nx, ny, nz = self._grid_creator.pillars.shape[-3:]
                out.create_dataset(
                    target, data=np.zeros((nz, nx, ny), dtype=np.float32), compression='gzip'
                )
            else:
                out.copy(self._archel_file[source], target)

        # Architectural elements data set (zero-array if key does not exist)
        _copy_archel_data(self._delft3d_archel_key, self._resqml_archel_key)

        # Sub-environment data set (zero-array if key does not exist)
        _copy_archel_data(self._delft3d_subenv_key, self._resqml_subenv_key)

        # Dump data
        self._grid_creator.dump_grid_data(out)

    def h5_base_name(self) -> str:
        return 'Delft3d.h5'



import argparse
import nrresqml
from time import perf_counter


parser = argparse.ArgumentParser(prog='python -m nrresqml', description='Convert Delft3D output to ResQml')
parser.add_argument(
    'delft3d_file', metavar='<delft-3d-file>', help='Input file generated by Delft3D'
)
parser.add_argument(
    'resqml_directory', metavar='<resqml_directory>', help='Destination directory for the ResQml output'
)

args = parser.parse_args()
t0 = perf_counter()
nrresqml.convert_delft3d_to_resqml(args.delft3d_file, args.resqml_directory)
t1 = perf_counter()
print(f'Conversion completed in {t1 - t0} s')



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple

from nrresqml.structures import xsd


class TargetMode(Enum):
    External = 0
    Undefined = 1


@dataclass
class Relationship(xsd.XsdComplexType):
    Id: str
    Target: str
    TargetMode: Optional[TargetMode]
    Type: str

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return []

    @classmethod
    def xml_attributes(cls) -> List[str]:
        return super().xml_attributes() + ['Id', 'Target', 'TargetMode', 'Type']


@dataclass
class Relationships(xsd.XsdComplexType):
    Relationship: List[Relationship]

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [(None, 'http://schemas.openxmlformats.org/package/2006/relationships')]



from enum import Enum
from typing import List, Tuple, Any, Optional
import datetime


xsi_uri = 'http://www.w3.org/2001/XMLSchema-instance'


class _GeneralType:
    @classmethod
    def main_namespace(cls, abbreviate) -> str:
        if len(cls.namespaces()) == 0:
            return ''
        ns = cls.namespaces()[0]
        return ns[0] if abbreviate else ns[1]

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        raise NotImplementedError('Method not implemented by subclass')

    @classmethod
    def type_name(cls) -> str:
        return cls.__name__

    @classmethod
    def full_type_name(cls, abbreviate_ns=False) -> str:
        if abbreviate_ns:
            prefix = f'{cls.main_namespace(abbreviate=True)}:'
        else:
            prefix = f'{{{cls.main_namespace(abbreviate=False)}}}'
        return f'{prefix}{cls.type_name()}'


class XsdComplexType(_GeneralType):
    @classmethod
    def content_type_string(cls) -> str:
        return f'application/x-resqml+xml;version=2.0.1;type=obj_{cls.type_name()}'

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [('xsi', 'http://www.w3.org/2001/XMLSchema-instance')]

    @classmethod
    def xml_attributes(cls) -> List[str]:
        """ Returns a list of names of attributes that will be added as XML attributes instead of nodes """
        return []


class BasicType(_GeneralType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        raise NotImplementedError('Method not implemented by subclass')


class BasicEnum(BasicType, Enum):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        raise NotImplementedError('Method not implemented by subclass')


class XsdBasicType(BasicType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [('xsi', 'http://www.w3.org/2001/XMLSchema-instance')]


class integer(XsdBasicType, int):
    pass


class nonNegativeInteger(integer):
    def __new__(cls, value) -> Any:
        value = int(value)
        assert value >= 0
        return super().__new__(cls, value)


class positiveInteger(integer):
    def __new__(cls, value) -> Any:
        value = int(value)
        assert value > 0
        return super().__new__(cls, value)


class double(XsdBasicType, float):
    pass


class string(XsdBasicType, str):
    pass


class boolean(XsdBasicType, int):
    def __new__(cls, value) -> Any:
        return super().__new__(cls, bool(value))


class dateTime(XsdBasicType, datetime.datetime):
    _fmt = "%Y-%m-%dT%H:%M:%SZ"

    def __new__(cls, *args, **kwargs) -> Any:
        if len(args) == 1 and isinstance(args[0], str):
            dt = datetime.datetime.strptime(args[0], cls._fmt)
            return super().__new__(cls, dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
        else:
            return super().__new__(cls, *args, **kwargs)

    def __str__(self) -> str:
        return self.strftime(self._fmt)



import typing
import inspect
from typing import Type, Optional

from lxml import etree

from nrresqml import reporting
from nrresqml.structures import relationships, contenttypes, xsd, energetics, core
from nrresqml.structures.resqml import properties, common, geometry, representations


_packages = [
    relationships,
    contenttypes,
    xsd,
    energetics,
    core,
    # ResQml specific
    properties,
    common,
    geometry,
    representations
]


_classes = {
    m: inspect.getmembers(m, inspect.isclass)
    for m in _packages
}


_xsi_type_uri = f'{{{xsd.xsi_uri}}}type'


class _ElementExtractValueError(Exception):
    pass


def _lookup_xsi_type(xsi_name: str) -> Optional[Type]:
    for module, clazzes in _classes.items():
        for clazz, type_ in clazzes:
            if type_.__module__ != module.__name__:
                continue
            try:
                ns = [f'{type_.main_namespace(abbreviate=True)}:',
                      f'{{{type_.main_namespace(abbreviate=False)}}}']
            except (AttributeError, NotImplementedError):
                ns = ['']
            if any(xsi_name == f'{n}{clazz}' for n in ns):
                return type_
    return None


def _determine_type(el: etree.Element) -> Optional[Type]:
    t = el.attrib.get(_xsi_type_uri)
    if t is None:
        raise NotImplementedError('')
    else:
        return _lookup_xsi_type(t)


def _iterate_attributes(type_: Type):
    for an, at in typing.get_type_hints(type_).items():
        yield an


def _find_child_values(el: etree.Element, tag_end: str) -> typing.List[typing.Any]:
    """
    Look up all values for a given tag on the provided etree.Element. However, the tag is interpreted as the end of a
    tag, which enabled look-up for tags while ignoring namespace if desired.
    """
    # Initialize 'out' with values from the element's attributes
    out = [value
           for key, value in el.attrib.items()
           # For attributes we ignore the namespace
           if tag_end.endswith(key)]
    # Append values from the element's children
    for child in el:
        if not child.tag.endswith(tag_end):
            continue
        if child.text is not None and child.text.strip() != '':
            txt = child.text.strip()
            ty = _determine_type(child)
            if issubclass(ty, xsd.BasicEnum):
                out.append(getattr(ty, txt))
            else:
                out.append(ty(txt))
        else:
            out.append(objectify(child))
    return out


def _is_optional(type_):
    orig = getattr(type_, '__origin__', None)
    return orig is typing.Union and isinstance(None, type_.__args__)


def _extract_values(el: etree.Element, att_name: str, att_type: Type
                    ) -> typing.Union[typing.List[typing.Any], typing.Any]:
    out = _find_child_values(el, att_name)
    # NB! This does not handle Optional[List]
    if getattr(att_type, '__origin__', None) is list:
        return out
    elif len(out) > 1:
        raise _ElementExtractValueError(f'Multiple elements found for tag {att_name}')
    elif _is_optional(att_type):
        return out
    elif len(out) == 0:
        raise _ElementExtractValueError(f'Failed to find tag {att_name}')
    else:
        return out[0]


def _iterate(el: etree.Element):
    for key, value in el.attrib.items():
        if key == _xsi_type_uri:
            continue
        yield key, value
    type_ = _determine_type(el)
    for an, at in typing.get_type_hints(type_).items():
        try:
            yield an, _extract_values(el, an, at)
        except _ElementExtractValueError as e:
            reporting.error(str(e))


def objectify(el: etree.Element):
    # TODO: avoid calling determine_type twice
    type_ = _determine_type(el)
    if type_ is None:
        reporting.error(f'Failed to determine type for XML tag {el.tag}')
        return
    ctor_dict = dict(_iterate(el))
    return type_(**ctor_dict)



import getpass
import uuid

from nrresqml.structures import xsd
from nrresqml.structures.energetics import Citation, UuidString, DescriptionString, NameString

_SCHEMA_VERSION = 'v2.0.1'


def create_meta_data(title):
    return dict(
        uuid=UuidString(str(uuid.uuid1())),
        schemaVersion=_SCHEMA_VERSION,
        Citation=Citation(
            Title=DescriptionString(title),
            Originator=NameString(getpass.getuser()),
            Format=DescriptionString('[NorwegianComputingCenter:netcdf2resqml]'),
            Creation=xsd.dateTime.now(),
        )
    )



from typing import List, Tuple, Optional

from nrresqml.structures.xsd import XsdComplexType, BasicEnum


_main_ns = ('resqml2', 'http://www.energistics.org/energyml/data/resqmlv2')


class ResqmlComplexType(XsdComplexType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns] + super().namespaces()


class ResqmlBasicEnum(BasicEnum):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns]



from setuptools import setup, find_packages


setup(
    name='nrresqml',
    version=open('nrresqml/VERSION.txt').read(),
    description='Package for converting Delft3D output to ResQml',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    author='Norwegian Computing Center',
    license='GPL-3.0',
    url='https://github.com/NorskRegnesentral/nrresqml',
    packages=find_packages(),
    package_data={
        'nrresqml': ['VERSION.txt']
    },
    # include_package_data=True,
    python_requires='>=3.7',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



import re
from dataclasses import dataclass
from typing import Optional, List, Tuple, Any
from nrresqml.structures import xsd


_main_ns = ('eml', 'http://www.energistics.org/energyml/data/commonv2')
_uuid_regex = re.compile(r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}")


class _EnergeticsEnum(xsd.BasicEnum):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns]


@dataclass
class _EnergeticsComplexType(xsd.XsdComplexType):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns] + super().namespaces()


""" Basic types """


class DescriptionString(xsd.string):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns] + super().namespaces()


class NameString(xsd.string):
    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns] + super().namespaces()


class UuidString(xsd.string):
    def __new__(cls, value) -> Any:
        assert re.match(_uuid_regex, value) is not None
        return super().__new__(cls, value)

    @classmethod
    def namespaces(cls) -> List[Tuple[Optional[str], str]]:
        return [_main_ns] + super().namespaces()


""" Complex types """


@dataclass
class DataObjectReference(xsd.XsdComplexType):
    ContentType: xsd.string
    Title: DescriptionString
    UUID: UuidString
    UuidAuthority: Optional[xsd.string]
    VersionString: Optional[NameString]


@dataclass
class Citation(_EnergeticsComplexType):
    Title: DescriptionString
    Originator: NameString
    Format: DescriptionString
    Creation: xsd.dateTime


@dataclass
class AbstractObject(_EnergeticsComplexType):
    schemaVersion: xsd.string
    uuid: UuidString
    Citation: Optional[Citation]

    @classmethod
    def xml_attributes(cls) -> List[str]:
        return super().xml_attributes() + ['uuid', 'schemaVersion']

    def base_name(self) -> str:
        return f'obj_{self.type_name()}_{self.uuid}.xml'


@dataclass
class AbstractCitedDataObject(AbstractObject):
    Citation: Citation


@dataclass
class EpcExternalPartReference(AbstractCitedDataObject):
    MimeType: xsd.string


@dataclass
class Hdf5Dataset(_EnergeticsComplexType):
    PathInHdfFile: xsd.string
    HdfProxy: EpcExternalPartReference


class AxisOrder2d(_EnergeticsEnum):
    easting_northing = 0
    northing_easting = 1
    westing_southing = 2
    southing_westing = 3
    northing_westing = 4
    westing_northing = 5


@dataclass
class AbstractVerticalCrs(_EnergeticsComplexType):
    pass


@dataclass
class AbstractProjectedCrs(_EnergeticsComplexType):
    pass


@dataclass
class VerticalUnknownCrs(AbstractVerticalCrs):
    Unknown: xsd.string


@dataclass
class ProjectedCrsEpsgCode(AbstractProjectedCrs):
    EpsgCode: xsd.positiveInteger


class LengthUom(_EnergeticsEnum):
    m = 0


class TimeUom(_EnergeticsEnum):
    ms = 0



from dataclasses import dataclass
from typing import Optional, List

from nrresqml.structures import xsd
from nrresqml.structures.resqml.geometry import AbstractValueArray
from nrresqml.structures.resqml.common import AbstractResqmlDataObject
from nrresqml.structures.resqml._rqtools import ResqmlComplexType, ResqmlBasicEnum
from nrresqml.structures.resqml.representations import IndexableElements, AbstractRepresentation


class ResqmlUom(ResqmlBasicEnum):
    m = 'm'


class ResqmlPropertyKind(ResqmlBasicEnum):
    discrete = 0
    porosity = 1
    permeability_rock = 2
    permeability_length = 3
    permeability_thickness = 4
    volume_per_volume = 5
    categorical = 6


""" PatchOfValues """


@dataclass
class PatchOfValues(ResqmlComplexType):
    RepresentationPatchIndex: xsd.integer
    Values: AbstractValueArray


""" AbstractPropertyKind """


@dataclass
class AbstractPropertyKind(ResqmlComplexType):
    pass


@dataclass
class PropertyKind(AbstractResqmlDataObject):
    NamingSystem: xsd.string
    IsAbstract: xsd.boolean
    Representative_uom: ResqmlUom
    ParentPropertyKind: AbstractPropertyKind


@dataclass
class LocalPropertyKind(AbstractPropertyKind):
    LocalPropertyKind: PropertyKind


@dataclass
class StandardPropertyKind(AbstractPropertyKind):
    Kind: ResqmlPropertyKind


""" AbstractPropertyLookup and StringLookup """


@dataclass
class StringLookup(ResqmlComplexType):
    Key: xsd.integer
    Value: xsd.string


@dataclass
class AbstractPropertyLookup(AbstractResqmlDataObject):
    pass


@dataclass
class StringTableLookup(AbstractPropertyLookup):
    Value: List[StringLookup]


""" AbstractProperty """


@dataclass
class AbstractProperty(AbstractResqmlDataObject):
    IndexableElement: IndexableElements
    Count: xsd.positiveInteger
    RealizationIndex: Optional[xsd.positiveInteger]
    TimeStep: Optional[xsd.positiveInteger]
    SupportingRepresentation: AbstractRepresentation
    PropertyKind: AbstractPropertyKind


@dataclass
class AbstractValuesProperty(AbstractProperty):
    PatchOfValues: PatchOfValues


@dataclass
class CategoricalProperty(AbstractValuesProperty):
    Lookup: AbstractPropertyLookup


@dataclass
class ContinuousProperty(AbstractValuesProperty):
    MinimumValue: xsd.double
    Uom: ResqmlUom  # underscores added since the true property name is UOM, deviating from CamelCase
    MaximumValue: xsd.double
